/* linux/arch/arm/mach-vt8500/gpio.c
 *
 * Copyright (C) 2010 Alexey Charkov <alchark@gmail.com>
 *
 * This software is licensed under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation, and
 * may be copied, distributed, and modified under those terms.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <linux/init.h>
#include <linux/irq.h>
#include <linux/io.h>

#include <mach/gpio.h>
#include <mach/irqs.h>
#include <mach/vt8500.h>

#define to_vt8500(__chip) container_of(__chip, struct vt8500_gpio_chip, chip)

static void __iomem *regbase;

struct vt8500_gpio_chip {
	struct gpio_chip	chip;
	unsigned int		shift;
	unsigned int		regoff;
};

static unsigned int gpio_to_irq_map[] = {
	IRQ_EXT0,
	IRQ_EXT1,
	IRQ_EXT2,
	IRQ_EXT3,
	IRQ_EXT4,
	IRQ_EXT5,
	IRQ_EXT6,
	IRQ_EXT7,
};

static int vt8500_muxed_gpio_request(struct gpio_chip *chip,
				     unsigned offset) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	writel(readl(regbase + vt8500_chip->regoff) |
		(1 << vt8500_chip->shift << offset),
		regbase + vt8500_chip->regoff);

	return 0;
}

static void vt8500_muxed_gpio_free(struct gpio_chip *chip,
				   unsigned offset) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	writel(readl(regbase + vt8500_chip->regoff) &
		~(1 << vt8500_chip->shift << offset),
		regbase + vt8500_chip->regoff);
}

static int vt8500_muxed_gpio_direction_input(struct gpio_chip *chip,
				       unsigned offset) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	writel(readl(regbase + 0x20 + vt8500_chip->regoff) &
		~(1 << vt8500_chip->shift << offset),
		regbase + 0x20 + vt8500_chip->regoff);

	return 0;
}

static int vt8500_muxed_gpio_direction_output(struct gpio_chip *chip,
					unsigned offset, int value) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	writel(readl(regbase + 0x20 + vt8500_chip->regoff) |
		(1 << vt8500_chip->shift << offset),
		regbase + 0x20 + vt8500_chip->regoff);

	if (value)
		writel(readl(regbase + 0x40 + vt8500_chip->regoff) |
			(1 << vt8500_chip->shift << offset),
			regbase + 0x40 + vt8500_chip->regoff);
	return 0;
}

static int vt8500_muxed_gpio_get_value(struct gpio_chip *chip,
				       unsigned offset) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	return (readl(regbase + 0x60 + vt8500_chip->regoff)
		>> vt8500_chip->shift >> offset) & 1;
}

static void vt8500_muxed_gpio_set_value(struct gpio_chip *chip,
					unsigned offset, int value) {
	struct vt8500_gpio_chip *vt8500_chip = to_vt8500(chip);

	if (value)
		writel(readl(regbase + 0x40 + vt8500_chip->regoff) |
			(1 << vt8500_chip->shift << offset),
			regbase + 0x40 + vt8500_chip->regoff);
	else
		writel(readl(regbase + 0x40 + vt8500_chip->regoff) &
			~(1 << vt8500_chip->shift << offset),
			regbase + 0x40 + vt8500_chip->regoff);
}

#define VT8500_GPIO_BANK(__name, __shift, __off, __base, __num)		\
{									\
	.chip = {							\
		.label			= __name,			\
		.request		= vt8500_muxed_gpio_request,	\
		.free			= vt8500_muxed_gpio_free,	\
		.direction_input  = vt8500_muxed_gpio_direction_input,	\
		.direction_output = vt8500_muxed_gpio_direction_output,	\
		.get			= vt8500_muxed_gpio_get_value,	\
		.set			= vt8500_muxed_gpio_set_value,	\
		.can_sleep		= 0,				\
		.base			= __base,			\
		.ngpio			= __num,			\
	},								\
	.shift		= __shift,					\
	.regoff		= __off,					\
}

static struct vt8500_gpio_chip vt8500_muxed_gpios[] = {
	VT8500_GPIO_BANK("uart0", 0, 0x0, 8, 4),
	VT8500_GPIO_BANK("uart1", 4, 0x0, 12, 4),
	VT8500_GPIO_BANK("spi0", 8, 0x0, 16, 4),
	VT8500_GPIO_BANK("spi1", 12, 0x0, 20, 4),
	VT8500_GPIO_BANK("spi2", 16, 0x0, 24, 4),
	VT8500_GPIO_BANK("pwmout", 24, 0x0, 28, 2),

	VT8500_GPIO_BANK("sdmmc", 0, 0x4, 30, 11),
	VT8500_GPIO_BANK("ms", 16, 0x4, 41, 7),
	VT8500_GPIO_BANK("i2c0", 24, 0x4, 48, 2),
	VT8500_GPIO_BANK("i2c1", 26, 0x4, 50, 2),

	VT8500_GPIO_BANK("mii", 0, 0x8, 52, 20),
	VT8500_GPIO_BANK("see", 20, 0x8, 72, 4),
	VT8500_GPIO_BANK("ide", 24, 0x8, 76, 7),

	VT8500_GPIO_BANK("ccir", 0, 0xc, 83, 19),

	VT8500_GPIO_BANK("ts", 8, 0x10, 102, 11),

	VT8500_GPIO_BANK("lcd", 0, 0x14, 113, 23),
};

static int vt8500_gpio_direction_input(struct gpio_chip *chip,
				       unsigned offset) {
	writel(readl(regbase + 0x3c) & ~(1 << offset), regbase + 0x3c);
	return 0;
}

static int vt8500_gpio_direction_output(struct gpio_chip *chip,
					unsigned offset, int value) {
	writel(readl(regbase + 0x3c) | (1 << offset), regbase + 0x3c);

	if (value)
		writel(readl(regbase + 0x5c) | (1 << offset),
		       regbase + 0x5c);
	return 0;
}

static int vt8500_gpio_get_value(struct gpio_chip *chip,
				       unsigned offset) {
	return (readl(regbase + 0x7c) >> offset) & 1;
}

static void vt8500_gpio_set_value(struct gpio_chip *chip,
					unsigned offset, int value) {
	if (value)
		writel(readl(regbase + 0x5c) | (1 << offset),
		       regbase + 0x5c);
	else
		writel(readl(regbase + 0x5c) & ~(1 << offset),
			regbase + 0x5c);
}

static int vt8500_gpio_to_irq(struct gpio_chip *chip, unsigned offset) {
	if (offset > 7)
		return -EINVAL;

	return gpio_to_irq_map[offset];
}

static struct gpio_chip vt8500_external_gpios = {
	.label			= "extgpio",
	.direction_input	= vt8500_gpio_direction_input,
	.direction_output	= vt8500_gpio_direction_output,
	.get			= vt8500_gpio_get_value,
	.set			= vt8500_gpio_set_value,
	.to_irq			= vt8500_gpio_to_irq,
	.can_sleep		= 0,
	.base			= 0,
	.ngpio			= 8,
};

void __init vt8500_gpio_init(void) {
	int i;

	regbase = ioremap(VT8500_GPIO_BASE, SZ_64K);

	gpiochip_add(&vt8500_external_gpios);

	for (i = 0; i < ARRAY_SIZE(vt8500_muxed_gpios); i++)
		gpiochip_add(&vt8500_muxed_gpios[i].chip);
}
