/*
 *  arch/arm/mach-vt8500/irq.c
 *
 *  Copyright (C) 2010 Alexey Charkov <alchark@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/io.h>
#include <linux/irq.h>
#include <linux/interrupt.h>

#include <asm/irq.h>

#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/vt8500.h>

#define VT8500_IC_DCTR		0x40		/* Destination control
						register, 64*u8 */
#define VT8500_INT_ENABLE	(1 << 3)
#define VT8500_TRIGGER_HIGH	(0 << 4)
#define VT8500_TRIGGER_RISING	(1 << 4)
#define VT8500_TRIGGER_FALLING	(2 << 4)
#define VT8500_IC_STATUS	0x80		/* Interrupt status, 2*u32 */

static void __iomem *ic_regbase = NULL;

static void vt8500_irq_mask(unsigned int irq)
{
	u8 edge = readb(ic_regbase
			+ VT8500_IC_DCTR + irq) & (3 << 4);
	if (edge)
		writel(readl(ic_regbase
			+ VT8500_IC_STATUS + (irq < 32 ? 0 : 4))
			| (1 << (irq & 0x1f)), ic_regbase
			+ VT8500_IC_STATUS + (irq & 0x20 ? 4 : 0));
	else
		writeb(readb(ic_regbase
			+ VT8500_IC_DCTR + irq) & ~VT8500_INT_ENABLE,
			ic_regbase + VT8500_IC_DCTR + irq);
}

static void vt8500_irq_unmask(unsigned int irq)
{
	writeb(readb(ic_regbase
		+ VT8500_IC_DCTR + irq) | VT8500_INT_ENABLE,
		ic_regbase + VT8500_IC_DCTR + irq);
}

static int vt8500_irq_set_wake(unsigned int irq, unsigned int on)
{
	return -EINVAL;
}

static int vt8500_irq_set_type(unsigned int irq, unsigned int flow_type)
{
	switch (flow_type) {
	case IRQF_TRIGGER_LOW:
		return -EINVAL;
	case IRQF_TRIGGER_HIGH:
		writeb((readb(ic_regbase
			+ VT8500_IC_DCTR + irq) & ~(3 << 4))
			| VT8500_TRIGGER_HIGH, ic_regbase
			+ VT8500_IC_DCTR + irq);
		set_irq_handler(irq, handle_level_irq);
		break;
	case IRQF_TRIGGER_FALLING:
		writeb((readb(ic_regbase
			+ VT8500_IC_DCTR + irq) & ~(3 << 4))
			| VT8500_TRIGGER_FALLING, ic_regbase
			+ VT8500_IC_DCTR + irq);
		set_irq_handler(irq, handle_edge_irq);
		break;
	case IRQF_TRIGGER_RISING:
		writeb((readb(ic_regbase
			+ VT8500_IC_DCTR + irq) & ~(3 << 4))
			| VT8500_TRIGGER_RISING, ic_regbase
			+ VT8500_IC_DCTR + irq);
		set_irq_handler(irq, handle_edge_irq);
		break;
	}

	return 0;
}

static struct irq_chip vt8500_irq_chip = {
	.name      = "vt8500",
	.ack       = vt8500_irq_mask,
	.mask      = vt8500_irq_mask,
	.unmask    = vt8500_irq_unmask,
	.set_wake  = vt8500_irq_set_wake,
	.set_type  = vt8500_irq_set_type,
};

void __init vt8500_init_irq(void)
{
	unsigned int i;

	if (!ic_regbase)
		ic_regbase = ioremap(VT8500_IC_BASE, SZ_64K);

	/* Enable rotating priority for IRQ */
	writel((1 << 6), ic_regbase + 0x20);
	writel(0, ic_regbase + 0x24);

	for (i = 0; i < NR_IRQS; i++) {
		/* Disable all interrupts and route them to IRQ */
		writeb(0x00, ic_regbase + VT8500_IC_DCTR + i);

		set_irq_chip(i, &vt8500_irq_chip);
		set_irq_handler(i, handle_level_irq);
		set_irq_flags(i, IRQF_VALID);
	}
}
